@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-# In[14]:
+# In[4]:
 
 
 try:
@@ -9,13 +9,13 @@ try:
         from IPython.display import display, Javascript
         display(Javascript('Jupyter.notebook.kernel.execute("nb_name = \'" + Jupyter.notebook.notebook_name + "\'");'));
 except:
-        pass
+    pass
 
 
-# In[15]:
+# In[5]:
 
 
-version = '0.100'
+version = '0.104'
 
 import re
 import tempfile
@@ -33,7 +33,7 @@ import time
 
 # # Helper Functions
 
-# In[16]:
+# In[6]:
 
 
 def is_testing():
@@ -43,7 +43,7 @@ def is_testing():
         return False
 
 
-# In[17]:
+# In[7]:
 
 
 def isnotebook():
@@ -54,7 +54,7 @@ def isnotebook():
     return False
 
 
-# In[18]:
+# In[8]:
 
 
 def testPrint(*args):
@@ -64,7 +64,7 @@ def testPrint(*args):
 
 # # Inspect
 
-# In[19]:
+# In[9]:
 
 
 import inspect
@@ -76,7 +76,7 @@ def debugFunctionName():
     return inspect.stack()[1][3]
 
 
-# In[20]:
+# In[10]:
 
 
 debugGetFileName(), debugGetLineNumber(), debugFunctionName()
@@ -84,7 +84,7 @@ debugGetFileName(), debugGetLineNumber(), debugFunctionName()
 
 # # Shell
 
-# In[21]:
+# In[11]:
 
 
 def shell(cmd):
@@ -97,7 +97,7 @@ def shell(cmd):
     return out
 
 
-# In[22]:
+# In[12]:
 
 
 shell_template = """/bin/bash << EOF 2>&1
@@ -119,13 +119,13 @@ def bash(cmd, lineList=True):
 
 # # Timing
 
-# In[23]:
+# In[13]:
 
 
 timeSinceLastCallDict = {}
 
 
-# In[24]:
+# In[14]:
 
 
 def timeSinceLastCall(name):
@@ -138,14 +138,14 @@ def timeSinceLastCall(name):
     return name + ": " + time_s
 
 
-# In[25]:
+# In[15]:
 
 
 if is_testing():
     print(timeSinceLastCall('test marker'))
 
 
-# In[26]:
+# In[16]:
 
 
 import time
@@ -173,7 +173,7 @@ if is_testing():
 
 # # Requiremtns
 
-# In[27]:
+# In[17]:
 
 
 requirements = []
@@ -205,7 +205,7 @@ import yaml
 
 # # Strings
 
-# In[28]:
+# In[18]:
 
 
 def sFloat0(f): return "{:0.0f}".format(f)
@@ -235,7 +235,7 @@ if is_testing():
     print(sFloatSci5(1.234567))
 
 
-# In[29]:
+# In[19]:
 
 
 def strUuid4():
@@ -246,14 +246,14 @@ def strUuid4():
     return uuid
 
 
-# In[30]:
+# In[20]:
 
 
 if is_testing():
     print(strUuid4())
 
 
-# In[31]:
+# In[21]:
 
 
 def strRand(n, num=True, upperCase=True, lowerCase=True):
@@ -265,7 +265,7 @@ def strRand(n, num=True, upperCase=True, lowerCase=True):
     return ''.join(random.choice(pool) for _ in range(n))
 
 
-# In[32]:
+# In[22]:
 
 
 if is_testing():
@@ -274,7 +274,7 @@ if is_testing():
     print(strRand(20, upperCase=False, lowerCase=False))
 
 
-# In[33]:
+# In[23]:
 
 
 def strMd5Pickle4(obj):
@@ -284,14 +284,14 @@ def strMd5Pickle4(obj):
     return md5.hexdigest()
 
 
-# In[34]:
+# In[24]:
 
 
 if is_testing():
     print(strMd5Pickle4(['a', 'b']))
 
 
-# In[35]:
+# In[25]:
 
 
 def strMd5String(s):
@@ -300,7 +300,7 @@ def strMd5String(s):
     return md5.hexdigest()
 
 
-# In[36]:
+# In[26]:
 
 
 if is_testing():
@@ -309,7 +309,7 @@ if is_testing():
 
 # # Multiprocess / Thread
 
-# In[37]:
+# In[27]:
 
 
 def mapThread(func, args, n=8):
@@ -334,14 +334,14 @@ if is_testing():
 
 # # Lists
 
-# In[38]:
+# In[28]:
 
 
 def liChunks(l, chunk_size):
     return [l[i:i + chunk_size] for i in range(0, len(l), chunk_size)]
 
 
-# In[39]:
+# In[29]:
 
 
 if is_testing():
@@ -350,7 +350,7 @@ if is_testing():
     print(liChunks(['a', 'b', 'c', 'd', 'e', 'f'], 3))
 
 
-# In[40]:
+# In[30]:
 
 
 def liSplit(a, number_of_splits):
@@ -358,7 +358,7 @@ def liSplit(a, number_of_splits):
     return [a[start:end] for start, end in zip(idxs[:-1], idxs[1:])]
 
 
-# In[41]:
+# In[31]:
 
 
 if is_testing():
@@ -370,7 +370,7 @@ if is_testing():
     print(liSplit(['a', 'b', 'c', 'd', 'e'], 6))
 
 
-# In[42]:
+# In[32]:
 
 
 def liJoin(l):
@@ -380,14 +380,14 @@ def liJoin(l):
     return out
 
 
-# In[43]:
+# In[33]:
 
 
 if is_testing():
     print(liJoin([['a', 'b', 'c'], ['d', 'e', 'f']]))
 
 
-# In[44]:
+# In[34]:
 
 
 def liFlatten(l):
@@ -404,7 +404,7 @@ def liFlatten(l):
         return l
 
 
-# In[45]:
+# In[35]:
 
 
 if is_testing():
@@ -414,7 +414,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[46]:
+# In[36]:
 
 
 regex_cache = {}
@@ -424,35 +424,35 @@ def sListFilter(l, re_include=".*", re_exclude='^\b$'):
     return [e for e in l if regex_cache[re_include].search(e) and not regex_cache[re_exclude].search(e)]
 
 
-# In[47]:
+# In[37]:
 
 
 if is_testing():
     sListFilter(['good x', 'other good x', 'without the letter before y'])
 
 
-# In[48]:
+# In[38]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[49]:
+# In[39]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[50]:
+# In[40]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[51]:
+# In[41]:
 
 
 def lCsv(l, force_oneline=False):
@@ -468,21 +468,21 @@ def lCsv(l, force_oneline=False):
     return ", ".join([str(e) for e in l])
 
 
-# In[52]:
+# In[42]:
 
 
 if is_testing():
     print(lCsv(['a', 'b', 1, 3.14]))
 
 
-# In[53]:
+# In[43]:
 
 
 if is_testing():
     print(lCsv([['a', 'b'], 1, 3.14]))
 
 
-# In[54]:
+# In[44]:
 
 
 if is_testing():
@@ -491,7 +491,7 @@ if is_testing():
 
 # # Text files
 
-# In[55]:
+# In[45]:
 
 
 def txtread(path):
@@ -512,7 +512,7 @@ def txtwrite(path, txt):
             return f.write(txt)
 
 
-# In[56]:
+# In[46]:
 
 
 if is_testing():
@@ -521,21 +521,21 @@ if is_testing():
 
 # # YAML files
 
-# In[57]:
+# In[47]:
 
 
 def yamlwrite(path, obj):
     txtwrite(path, yaml.dump(obj))
 
 
-# In[58]:
+# In[48]:
 
 
 def yamlread(path):
     return yaml.safe_load(txtread(path=path))
 
 
-# In[59]:
+# In[49]:
 
 
 if is_testing():
@@ -548,7 +548,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[60]:
+# In[50]:
 
 
 regex_cache = {}
@@ -560,35 +560,35 @@ def sListFilter(listOfStrings, re_include=".*", re_exclude='^\b$'):
 
 # ### Examples
 
-# In[61]:
+# In[51]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y']))
 
 
-# In[62]:
+# In[52]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[63]:
+# In[53]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[64]:
+# In[54]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[65]:
+# In[55]:
 
 
 if is_testing():
@@ -597,7 +597,7 @@ if is_testing():
 
 # # GCP Google Cloud Platform
 
-# In[66]:
+# In[56]:
 
 
 import warnings
@@ -621,7 +621,7 @@ def _gsBlob(path):
     return blob
 
 
-# In[67]:
+# In[57]:
 
 
 def gsGsToHttp(path):
@@ -631,7 +631,7 @@ def gsHttpToGs(path):
     return path.replace("https://storage.cloud.google.com/", "gs://").split("?")[0] # "?" is for possible "?authuser=0"
 
 
-# In[68]:
+# In[58]:
 
 
 def gsLs(path):
@@ -641,7 +641,7 @@ def gsLs(path):
     return ['gs://' + bucket + '/' + blob.name for blob in client.list_blobs(prefix=prefix)]
 
 
-# In[69]:
+# In[59]:
 
 
 def gsExists(path):
@@ -651,7 +651,7 @@ def gsExists(path):
     return client.blob(prefix).exists()
 
 
-# In[70]:
+# In[60]:
 
 
 def gsUploadFile(gsPath, path):
@@ -659,7 +659,7 @@ def gsUploadFile(gsPath, path):
     blob.upload_from_filename(path)
 
 
-# In[71]:
+# In[61]:
 
 
 def gsTxtread(path):
@@ -671,7 +671,7 @@ def gsTxtread(path):
     return blob.download_as_string().decode()
 
 
-# In[72]:
+# In[62]:
 
 
 def gsTxtwrite(path, s):
@@ -679,7 +679,7 @@ def gsTxtwrite(path, s):
     return blob.upload_from_string(s)
 
 
-# In[73]:
+# In[63]:
 
 
 def gsImwrite(path, img):
@@ -688,7 +688,7 @@ def gsImwrite(path, img):
     blob.upload_from_string(encimg.tobytes())
 
 
-# In[74]:
+# In[64]:
 
 
 def gsImread(path):
@@ -697,7 +697,7 @@ def gsImread(path):
     return cv2.imdecode(raw, 1)[:, :, [2, 1, 0]]
 
 
-# In[75]:
+# In[65]:
 
 
 def gsPklWrite(path, obj):
@@ -705,7 +705,7 @@ def gsPklWrite(path, obj):
     return blob.upload_from_string(pickle.dumps(obj, protocol=4))
 
 
-# In[76]:
+# In[66]:
 
 
 def gsPklRead(path):
@@ -715,7 +715,7 @@ def gsPklRead(path):
 
 # # Pickle
 
-# In[77]:
+# In[67]:
 
 
 def pklwrite(path, obj):
@@ -728,7 +728,7 @@ def pklwrite(path, obj):
             return pickle.dump(obj, f, protocol=4)
 
 
-# In[78]:
+# In[68]:
 
 
 def pklread(path):
@@ -739,7 +739,7 @@ def pklread(path):
             return pickle.load(f)
 
 
-# In[79]:
+# In[69]:
 
 
 if is_testing():
@@ -752,7 +752,7 @@ if is_testing():
 
 # # Images
 
-# In[80]:
+# In[70]:
 
 
 def _imreadSsh(path):
@@ -789,18 +789,18 @@ def imwrite(path, img, jpeg_quality=95):
     os.makedirs(os.path.dirname(path), exist_ok=True)
     
     if path[-4:] == '.jpg' or path[-4:] == '.jpeg':
-        cv2.imwrite(path, img, [int(cv2.IMWRITE_JPEG_QUALITY), jpeg_quality])
+        cv2.imwrite(path, img[:, :, [2, 1, 0]], [int(cv2.IMWRITE_JPEG_QUALITY), jpeg_quality])
     else:
         cv2.imwrite(path, img[:, :, [2, 1, 0]])
 
 def imjoin(imgs, axis=1, pad=0):
     if pad > 0:
         if axis == 0:
-            right=pad
-            bottom=0
-        elif axis == 1:
             right=0
             bottom=pad
+        elif axis == 1:
+            right=pad
+            bottom=0
         imgs_out = []
         for idx, img in enumerate(imgs):
             if idx < len(imgs) - 1:
@@ -819,7 +819,7 @@ def impad(img, top=0, bottom=0, left=0, right=0, color=255, mode='constant'):
         raise RuntimeError(f"Mode '{mode}' not found")
 
 
-# In[81]:
+# In[71]:
 
 
 def imscaleNN(img, s):
@@ -831,7 +831,7 @@ def imscaleBic(img, s):
 
 
 
-# In[82]:
+# In[72]:
 
 
 def imshow(array, scale=1, force_nb=False):
@@ -846,14 +846,14 @@ def imshow(array, scale=1, force_nb=False):
         cv2.waitKey()
 
 
-# In[83]:
+# In[73]:
 
 
 def imNewNoise(height, width, scaleNN=1):
     return imscaleNN((255 * np.random.rand(height, width, 3)).astype(np.uint8), scaleNN)
 
 
-# In[84]:
+# In[74]:
 
 
 if is_testing():
@@ -861,7 +861,7 @@ if is_testing():
     imshow(impad(img, bottom=10, right=10, mode='reflect'))
 
 
-# In[85]:
+# In[75]:
 
 
 def imCropUniformBorder(img, color=255):
@@ -900,7 +900,7 @@ if is_testing():
                         raise RuntimeError('Crop did not work for ', i, j, k, l)
 
 
-# In[86]:
+# In[76]:
 
 
 def imDecodeBytes(imBytes):
@@ -911,7 +911,7 @@ def imDecodeBytes(imBytes):
     return img
 
 
-# In[87]:
+# In[77]:
 
 
 #if is_testing():
@@ -919,14 +919,14 @@ def imDecodeBytes(imBytes):
 #    imshow(img)
 
 
-# In[88]:
+# In[78]:
 
 
 #if is_testing():
 #    imshow('http://via.placeholder.com/70.png')
 
 
-# In[89]:
+# In[79]:
 
 
 if is_testing():
@@ -945,7 +945,7 @@ if is_testing():
         imshow(out_large_interbic)
 
 
-# In[90]:
+# In[80]:
 
 
 def imRepeat(img, height, width):
@@ -959,7 +959,7 @@ def imRepeat(img, height, width):
     return img
 
 
-# In[91]:
+# In[81]:
 
 
 if is_testing():
@@ -967,33 +967,33 @@ if is_testing():
     imshow(imRepeat(img_random_color_3x5, 7, 11), scale=10)
 
 
-# In[92]:
+# In[82]:
 
 
 def imNewWhite(height, width):
     return np.ones((height, width, 3)).astype(np.uint8) * 255
 
 
-# In[93]:
+# In[83]:
 
 
 if is_testing(): imshow(imNewWhite(10, 500) - 30)
 
 
-# In[94]:
+# In[84]:
 
 
 def imNewBlack(height, width):
     return np.zeros((height, width, 3)).astype(np.uint8)
 
 
-# In[95]:
+# In[85]:
 
 
 if is_testing(): imshow(imNewBlack(10, 500))
 
 
-# In[96]:
+# In[86]:
 
 
 def imAddNoiseGauss(img, std):
@@ -1005,7 +1005,7 @@ def imAddNoiseGauss(img, std):
 if is_testing(): imshow(imAddNoiseGauss(imNewWhite(150, 150), 30))
 
 
-# In[97]:
+# In[87]:
 
 
 if is_testing():
@@ -1013,7 +1013,7 @@ if is_testing():
     imshow('/tmp/img.jpg')
 
 
-# In[98]:
+# In[88]:
 
 
 def imJpgDegradation(img, quality):
@@ -1026,7 +1026,7 @@ def imJpgDegradation(img, quality):
 #if is_testing(): imshow(imJpgDegradation(imread('http://via.placeholder.com/150.png'), 15))
 
 
-# In[99]:
+# In[89]:
 
 
 def imCropCenter(img, size):
@@ -1041,35 +1041,35 @@ def imCropCenter(img, size):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[100]:
+# In[90]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 10))
 
 
-# In[101]:
+# In[91]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 49))
 
 
-# In[102]:
+# In[92]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 50))
 
 
-# In[103]:
+# In[93]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 100))
 
 
-# In[104]:
+# In[94]:
 
 
 def imCropRandom(img, size, auto_size=True):
@@ -1092,7 +1092,7 @@ def imCropRandom(img, size, auto_size=True):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[105]:
+# In[95]:
 
 
 #if is_testing(): 
@@ -1102,7 +1102,95 @@ def imCropRandom(img, size, auto_size=True):
 #    imshow(imCropRandom(imread('http://via.placeholder.com/40.png'), 41))
 
 
-# In[106]:
+# In[96]:
+
+
+def draw_rect(img, h_beg, h_end, w_beg, w_end, line_width, color=(0, 0, 0)):
+    h, w, _ = img.shape
+    h_beg = max(h_beg, line_width)
+    w_beg = max(w_beg, line_width)
+    h_end = min(h_end, h - line_width)
+    w_end = min(w_end, w - line_width)
+
+    # top
+    img[h_beg - line_width: h_beg, w_beg - line_width:w_end + line_width] = color
+    # bottom
+    img[h_end: h_end + line_width, w_beg - line_width:w_end + line_width] = color
+    # left
+    img[h_beg - line_width:h_end + line_width, w_beg - line_width: w_beg] = color
+    # right
+    img[h_beg - line_width:h_end + line_width, w_end: w_end + line_width] = color
+
+    return img
+
+
+def imCropRandomSeed(img, sizes, idx, crop_idx, use_crop=True, use_minimap=False, mark_color=(200, 50, 50), scale=2, minimap_region=None):
+    
+    def get_beg_end(beg_crop, size_crop, size_region, size_img):
+        center = beg_crop + (size_crop // 2)
+        beg = center - int(round(size_region / 2))
+        if beg < 0:
+            beg = 0
+            end = size_region
+            return beg, end
+        end = center + int(round(size_region / 2))
+        if end > size_img:
+            end = size_img
+            beg = size_img - size_region
+            return beg, end
+        return beg, end
+
+    randState = np.random.get_state()
+
+    np.random.seed(idx)
+
+    h, w, _ = img.shape
+    for i in range(crop_idx + 1):
+        h_start = np.random.randint(0, h - sizes[0])
+        w_start = np.random.randint(0, w - sizes[1])
+
+    h_end = h_start + sizes[0]
+    w_end = w_start + sizes[1]
+
+    np.random.set_state(randState)
+
+    crop = img[h_start:h_end, w_start:w_end]
+    crop = imscaleNN(crop, scale)
+
+    if use_minimap:
+        img = img.copy()
+        img = draw_rect(img, h_start, h_end, w_start, w_end, line_width=10, color=mark_color)
+
+        if minimap_region is not None:
+            h_new = sizes[0] * minimap_region
+            w_new = sizes[1] * minimap_region
+            print(minimap_region)
+            h_beg, h_end = get_beg_end(h_start, sizes[0], h_new, h)
+            w_beg, w_end = get_beg_end(w_start, sizes[1], w_new, w)
+            img = img[h_beg:h_end, w_beg:w_end]
+            minimap = cv2.resize(img, (sizes[0] * scale, sizes[1] * scale))
+        else:
+            h_new = sizes[0] * scale
+            w_new = int(round(h_new * w / h))
+            minimap = cv2.resize(img, (w_new, h_new))
+
+        if not use_crop:
+            return minimap
+        else:
+            return imjoin([crop, minimap], pad=5)
+    else:
+        return crop
+
+
+# In[97]:
+
+
+if is_testing():
+    img = imscaleNN(imNewNoise(10, 20), 10)
+    imshow(imCropRandomSeed(img, (20, 20), idx=0, crop_idx=0, use_crop=True, use_minimap=True), scale=3)
+
+
+# In[98]:
 
 
 def imGallery(imgs, pad=0, n_cols=None, n_rows=None):
@@ -1146,7 +1234,7 @@ def imGallery(imgs, pad=0, n_cols=None, n_rows=None):
     return out
 
 
-# In[107]:
+# In[99]:
 
 
 if is_testing():
@@ -1154,7 +1242,7 @@ if is_testing():
     listOfDummyImagesBlackToWhite = [imNewBlack(10, 10) + 25 * i for i in range(10)]
 
 
-# In[108]:
+# In[100]:
 
 
 if is_testing():
@@ -1168,7 +1256,7 @@ if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1, n_rows=2))
 
 
-# In[109]:
+# In[101]:
 
 
 #if is_testing():
@@ -1176,7 +1264,7 @@ if is_testing():
 #    imshow(imGallery([img40x20, img40x20, img40x20], pad=1))
 
 
-# In[110]:
+# In[102]:
 
 
 def imAddTextHeader(img, text, h=None, size=None):
@@ -1213,7 +1301,7 @@ if is_testing():
 
 # # Video
 
-# In[111]:
+# In[103]:
 
 
 #def imgsToVid(imgs, file_name='out.avi'):
@@ -1233,13 +1321,13 @@ if is_testing():
 #imgsToVid(frames, 'test.avi')
 
 
-# In[112]:
+# In[104]:
 
 
 #ffmpeg -framerate 1/5 -i img%03d.png -c:v libx264 -r 30 -pix_fmt yuv420p out.mp4
 
 
-# In[113]:
+# In[105]:
 
 
 def imsToMp4(path, imgs, time_per_image, frame_rate_output_video=30, verbose=0):
@@ -1259,7 +1347,7 @@ if is_testing():
 
 # # pytorch
 
-# In[114]:
+# In[106]:
 
 
 def ptT(img_np_uint8):
@@ -1276,7 +1364,7 @@ def ptUint8(img_tensor):
     
 
 
-# In[115]:
+# In[107]:
 
 
 if is_testing():
@@ -1288,7 +1376,7 @@ if is_testing():
 
 # # Files
 
-# In[116]:
+# In[108]:
 
 
 def _fiFindByWildcardSsh(wildcard):
@@ -1309,13 +1397,16 @@ def fiFindImgs(dir_path, recursive=False, ext='png'):
     return fiFindByWildcard(wildcard)
 
 
-# In[117]:
+# In[109]:
 
 
 import os
 from os.path import join as pjoin
 from os.path import dirname as pdirname
 from os.path import basename as pfilename
+from os.path import isfile as pisfile
+from os.path import isdir as pisdir
+from os.path import expanduser as pexpanduser
 def pfilenameNoExt(p):
     splits = pfilename(p).split('.')
     assert len(splits) <= 2
@@ -1326,9 +1417,10 @@ if is_testing():
     print(pdirname('/tmp/file.txt'))
     print(pfilename('/tmp/file.txt'))
     print(pfilenameNoExt('/tmp/file.txt'))
+    print(pexpanduser('~/tmp/file.txt'))
 
 
-# In[118]:
+# In[110]:
 
 
 if is_testing():
@@ -1357,7 +1449,7 @@ if is_testing():
         imshow(imGallery(imgs))
 
 
-# In[119]:
+# In[111]:
 
 
 def fiRemove(path):
@@ -1385,7 +1477,7 @@ def fiAgeHours(path):
     return (time.time() - os.stat(path)[stat.ST_MTIME]) / 3600
 
 
-# In[120]:
+# In[112]:
 
 
 if is_testing():
@@ -1417,7 +1509,7 @@ if is_testing():
         print("files: " + "\n".join(fiFindByWildcard(os.path.join(d, "**", "*"))))
 
 
-# In[121]:
+# In[113]:
 
 
 def fiUnzip(zip_path, out_dir=None, remove=False):
@@ -1432,7 +1524,7 @@ def fiUnzip(zip_path, out_dir=None, remove=False):
         zip.extractall(out_dir)
 
 
-# In[122]:
+# In[114]:
 
 
 def fiJoin(*args):
@@ -1444,7 +1536,7 @@ if is_testing():
 
 # # Jupyter
 
-# In[123]:
+# In[115]:
 
 
 def jnSettings():
@@ -1455,7 +1547,7 @@ if is_testing():
     jnSettings()
 
 
-# In[124]:
+# In[116]:
 
 
 def jnNbFullWidth():
@@ -1469,7 +1561,7 @@ def jnPandasSettings():
     pd.set_option('display.max_colwidth', 100000)
 
 
-# In[125]:
+# In[117]:
 
 
 def jnImageRenderingPixelated():
@@ -1481,7 +1573,7 @@ def jnImageRenderingAuto():
     display(HTML('<style type = text/css> img { image-rendering: auto; } </style>'))
 
 
-# In[126]:
+# In[118]:
 
 
 def jnDefault():
@@ -1499,7 +1591,7 @@ if is_testing():
 
 
 
-# In[127]:
+# In[119]:
 
 
 def jnHtmlRow(strList, width=100):
@@ -1512,7 +1604,7 @@ if is_testing():
     
 
 
-# In[128]:
+# In[120]:
 
 
 def jnImgToHtmlB64(img):
@@ -1547,7 +1639,7 @@ def jnImageRow(imgs, names=None, width=100, external=False, title_img=None, titl
     display(HTML(html))
 
 
-# In[129]:
+# In[121]:
 
 
 if is_testing():
@@ -1555,7 +1647,7 @@ if is_testing():
    
 
 
-# In[130]:
+# In[122]:
 
 
 if is_testing():
@@ -1565,7 +1657,7 @@ if is_testing():
     jnImageRow(imgs, external=True, title='test')
 
 
-# In[131]:
+# In[123]:
 
 
 
@@ -1573,9 +1665,23 @@ if is_testing():
    jnImageRow(imgs=[imNewBlack(20, 20), imNewBlack(40, 20), imNewBlack(20, 800)])
 
 
+# In[124]:
+
+
+#def dfDisplayWithImgs(df):
+#    html = df.to_html()
+#    urls = list(set([m.group() for m in re.finditer('http://[-\w/_\.]*\.(jpg|jpeg|png)', html)]))
+#    html_out = html
+#    for url in urls:
+#        url_new = f"<img src='{url}'>"
+#        html_out = html_out.replace(url, url_new)
+#        assert url_new in html_out, url_ne
+#    display(HTML(html_out))
+
+
 # # GIT
 
-# In[132]:
+# In[125]:
 
 
 def gitGetHash():
@@ -1585,7 +1691,7 @@ if is_testing():
     print(gitGetHash())
 
 
-# In[133]:
+# In[126]:
 
 
 def gitGetBranch():
@@ -1595,7 +1701,7 @@ if is_testing():
     print(gitGetBranch())
 
 
-# In[134]:
+# In[127]:
 
 
 def gitGetMessage():
@@ -1605,7 +1711,7 @@ if is_testing():
     print(gitGetMessage())
 
 
-# In[135]:
+# In[128]:
 
 
 def gitGetAllFiles():
@@ -1615,7 +1721,7 @@ if is_testing():
     print(gitGetAllFiles())
 
 
-# In[136]:
+# In[129]:
 
 
 def gitListByTime(path=None):
@@ -1631,7 +1737,7 @@ if is_testing():
     print(gitListByTime()[-3:])
 
 
-# In[137]:
+# In[130]:
 
 
 def gitTime(path):
@@ -1642,7 +1748,7 @@ if is_testing():
     print(gitTime("__init__.py"))
 
 
-# In[138]:
+# In[131]:
 
 
 def gitTimeRaw(path):
@@ -1653,7 +1759,7 @@ if is_testing():
     print(gitTimeRaw("__init__.py"))
 
 
-# In[139]:
+# In[132]:
 
 
 def shellSortDirsBySize():
@@ -1665,14 +1771,14 @@ if is_testing():
 
 # # Package Files
 
-# In[140]:
+# In[133]:
 
 
 if is_testing():
     package_files = {}
 
 
-# In[141]:
+# In[134]:
 
 
 def shell(cmd):
@@ -1680,7 +1786,7 @@ def shell(cmd):
     return result.stdout.decode()
 
 
-# In[142]:
+# In[135]:
 
 
 if is_testing():
@@ -1712,7 +1818,7 @@ setup(
 )""".format(requirements=", ".join(["'" + r + "'" for r in requirements]), version=version)
 
 
-# In[143]:
+# In[136]:
 
 
 if is_testing():
@@ -1722,7 +1828,7 @@ description-file = README.md
 """
 
 
-# In[144]:
+# In[137]:
 
 
 if is_testing():
@@ -1745,7 +1851,7 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE."""
 
 
-# In[145]:
+# In[138]:
 
 
 if is_testing():
@@ -1763,7 +1869,7 @@ if is_testing():
     print(output.strip(), 'Version correct: ', output.strip() == version, "(", output.strip(), version, ")")
 
 
-# In[147]:
+# In[ ]:
 
 
 if is_testing():
@@ -1774,7 +1880,7 @@ lx/__init__.py
 """
 
 
-# In[148]:
+# In[ ]:
 
 
 if is_testing():
